//FCUP, VFS 2019/20

// Finds the maximum value in a non-empty array.
method findMax(a: array<real>) returns (max: real)
  requires a.Length > 0
  ensures exists k :: 0 <= k < a.Length && max == a[k]
  ensures forall k :: 0 <= k < a.Length ==> max >= a[k]
{
    max := a[0];
    var i := 1;  
    while i  <  a.Length
      decreases a.Length - i
      invariant 0 <= i <= a.Length
      invariant exists k :: 0 <= k < i && max == a[k]
      invariant forall k :: 0 <= k < i ==> max >= a[k]
    {
        if (a[i] > max)
        {
            max := a[i];
        }
        i := i+1;
    } 
    return max;
}

// A simple test case to check on the console
method Main()
{
    var a := new real[3];
    a[0], a[1], a[2] := 1.0, 2.0, 2.0;
    var max := findMax(a);
    assert max == a[1] || max == a[2]; // trick to accept next ...
    assert max == 2.0;
    print "max(1.0, 2.0, 3.0) = ";
    print max;
    print "\n";
}

// Series of test cases verified statically ...
method testMaxSortedAsc()
{
    var a := new real[3];
    a[0], a[1], a[2] := 1.0, 2.0, 3.0;
    var m := findMax(a);
    assert m == a[2] == 3.0;
}

method testMaxSortedDesc()
{
    var a := new real[3];
    a[0], a[1], a[2] := 3.0, 2.0, 1.0;
    var m := findMax(a);
    assert m == a[0] == 3.0;
}

method testMaxUnsorted()
{
    var a := new real[3];
    a[0], a[1], a[2] := 2.0, 3.0, 1.0;
    var m := findMax(a);
    assert m == a[1] == 3.0;
}

method testMaxDuplicates()
{
    var a := new real[3];
    a[0], a[1], a[2] := 1.0, 2.0, 2.0;
    var m := findMax(a);
    assert a[1] == 2.0;
    assert m == 2.0; 
}

